from ..bbot_fixtures import *


def test_python_api(bbot_config):
    from bbot.scanner import Scanner

    # make sure events are properly yielded
    scan1 = Scanner("127.0.0.1", config=bbot_config)
    events1 = list(scan1.start())
    assert any("127.0.0.1" == e for e in events1)
    # make sure output files work
    scan2 = Scanner("127.0.0.1", config=bbot_config, output_modules=["json"], name="python_api_test")
    scan2.start_without_generator()
    out_file = scan2.helpers.scans_dir / "python_api_test" / "output.json"
    assert list(scan2.helpers.read_file(out_file))
    # make sure config loads properly
    bbot_home = "/tmp/.bbot_python_api_test"
    Scanner("127.0.0.1", config={"home": bbot_home})
    assert os.environ["BBOT_TOOLS"] == str(Path(bbot_home) / "tools")



import re
import json
import logging
import ipaddress
import traceback
import cloudcheck
import dns.resolver
import dns.exception
from threading import Lock
from contextlib import suppress
from concurrent.futures import ThreadPoolExecutor

from .threadpool import NamedLock
from .regexes import dns_name_regex
from bbot.core.errors import ValidationError, DNSError
from .misc import is_ip, is_domain, domain_parents, parent_domain, rand_string

log = logging.getLogger("bbot.core.helpers.dns")


class DNSHelper:
    """
    For automatic wildcard detection, nameserver validation, etc.
    """

    nameservers_url = "https://public-dns.info/nameserver/nameservers.json"
    all_rdtypes = ["A", "AAAA", "SRV", "MX", "NS", "SOA", "CNAME", "TXT"]

    def __init__(self, parent_helper):
        self.parent_helper = parent_helper
        try:
            self.resolver = dns.resolver.Resolver()
        except Exception as e:
            raise DNSError(f"Failed to create BBOT DNS resolver: {e}")
        self.timeout = self.parent_helper.config.get("dns_timeout", 5)
        self.retries = self.parent_helper.config.get("dns_retries", 1)
        self.abort_threshold = self.parent_helper.config.get("dns_abort_threshold", 5)
        self.max_dns_resolve_distance = self.parent_helper.config.get("max_dns_resolve_distance", 4)
        self.resolver.timeout = self.timeout
        self.resolver.lifetime = self.timeout
        self._resolver_list = None

        self.wildcard_ignore = self.parent_helper.config.get("dns_wildcard_ignore", None)
        if not self.wildcard_ignore:
            self.wildcard_ignore = []
        self.wildcard_ignore = tuple([str(d).strip().lower() for d in self.wildcard_ignore])
        self.wildcard_tests = self.parent_helper.config.get("dns_wildcard_tests", 5)
        self._wildcard_cache = dict()
        # since wildcard detection takes some time, This is to prevent multiple
        # modules from kicking off wildcard detection for the same domain at the same time
        self._wildcard_lock = NamedLock()
        # keeps track of warnings issued for wildcard detection to prevent duplicate warnings
        self._dns_warnings = set()

        self._errors = dict()
        self._error_lock = Lock()

        self.fallback_nameservers_file = self.parent_helper.wordlist_dir / "nameservers.txt"

        # we need our own threadpool because using the shared one can lead to deadlocks
        max_workers = self.parent_helper.config.get("max_dns_threads", 100)
        self._thread_pool = ThreadPoolExecutor(max_workers=max_workers)

        self._debug = self.parent_helper.config.get("dns_debug", False)

        self._dummy_modules = dict()
        self._dummy_modules_lock = Lock()

        self._dns_cache = self.parent_helper.CacheDict(max_size=100000)

        self._event_cache = self.parent_helper.CacheDict(max_size=10000)
        self._event_cache_lock = Lock()
        self._event_cache_locks = NamedLock()

        # copy the system's current resolvers to a text file for tool use
        self.system_resolvers = dns.resolver.Resolver().nameservers
        self.resolver_file = self.parent_helper.tempfile(self.system_resolvers, pipe=False)

        self.bad_ptr_regex = re.compile(r"(?:[0-9]{1,3}[-_\.]){3}[0-9]{1,3}")
        self.filter_bad_ptrs = self.parent_helper.config.get("dns_filter_ptrs", True)

    def resolve(self, query, **kwargs):
        """
        "1.2.3.4" --> {
            "evilcorp.com",
        }
        "evilcorp.com" --> {
            "1.2.3.4",
            "dead::beef"
        }
        """
        results = set()
        raw_results, errors = self.resolve_raw(query, **kwargs)
        for rdtype, answers in raw_results:
            for answer in answers:
                for _, t in self.extract_targets(answer):
                    results.add(t)
        return results

    def resolve_raw(self, query, **kwargs):
        # DNS over TCP is more reliable
        # But setting this breaks DNS resolution on Ubuntu because systemd-resolve doesn't support TCP
        # kwargs["tcp"] = True
        if self.parent_helper.scan.stopping:
            return [], []
        query = str(query).strip()
        if is_ip(query):
            kwargs.pop("type", None)
            kwargs.pop("rdtype", None)
            results, errors = self._resolve_ip(query, **kwargs)
            return [("PTR", results)], [("PTR", e) for e in errors]
        else:
            results = []
            errors = []
            types = ["A", "AAAA"]
            kwargs.pop("rdtype", None)
            if "type" in kwargs:
                t = kwargs.pop("type")
                if isinstance(t, str):
                    if t.strip().lower() in ("any", "all", "*"):
                        types = self.all_rdtypes
                    else:
                        types = [t.strip().upper()]
                elif any([isinstance(t, x) for x in (list, tuple)]):
                    types = [str(_).strip().upper() for _ in t]
            for t in types:
                if getattr(self.parent_helper.scan, "stopping", False) == True:
                    break
                r, e = self._resolve_hostname(query, rdtype=t, **kwargs)
                if r:
                    results.append((t, r))
                for error in e:
                    errors.append((t, error))

            return (results, errors)

    def submit_task(self, *args, **kwargs):
        try:
            return self._thread_pool.submit(*args, **kwargs)
        except RuntimeError as e:
            log.debug(f"Error submitting DNS thread task: {e}")

    def _resolve_hostname(self, query, **kwargs):
        self.debug(f"Resolving {query} with kwargs={kwargs}")
        results = []
        errors = []
        parent = self.parent_helper.parent_domain(query)
        rdtype = kwargs.get("rdtype", "A")
        retries = kwargs.pop("retries", self.retries)
        cache_result = kwargs.pop("cache_result", False)
        tries_left = int(retries) + 1
        parent_hash = hash(f"{parent}:{rdtype}")
        dns_cache_hash = hash(f"{query}:{rdtype}")
        while tries_left > 0:
            try:
                try:
                    results = self._dns_cache[dns_cache_hash]
                except KeyError:
                    error_count = self._errors.get(parent_hash, 0)
                    if error_count >= self.abort_threshold:
                        log.verbose(
                            f'Aborting query "{query}" because failed {rdtype} queries for "{parent}" ({error_count:,}) exceeded abort threshold ({self.abort_threshold:,})'
                        )
                        return results, errors
                    results = list(self._catch(self.resolver.resolve, query, **kwargs))
                    if cache_result:
                        self._dns_cache[dns_cache_hash] = results
                    with self._error_lock:
                        if parent_hash in self._errors:
                            self._errors[parent_hash] = 0
                break
            except (dns.resolver.NoNameservers, dns.exception.Timeout, dns.resolver.LifetimeTimeout) as e:
                with self._error_lock:
                    try:
                        self._errors[parent_hash] += 1
                    except KeyError:
                        self._errors[parent_hash] = 1
                log.verbose(
                    f'DNS error or timeout for {rdtype} query "{query}" ({self._errors[parent_hash]:,} so far): {e}'
                )
                errors.append(e)
                # don't retry if we get a SERVFAIL
                if isinstance(e, dns.resolver.NoNameservers):
                    break
                tries_left -= 1
                if tries_left > 0:
                    retry_num = (retries + 1) - tries_left
                    self.debug(f"Retry (#{retry_num}) resolving {query} with kwargs={kwargs}")

        self.debug(f"Results for {query} with kwargs={kwargs}: {results}")
        return results, errors

    def _resolve_ip(self, query, **kwargs):
        self.debug(f"Reverse-resolving {query} with kwargs={kwargs}")
        retries = kwargs.pop("retries", 0)
        cache_result = kwargs.pop("cache_result", False)
        tries_left = int(retries) + 1
        results = []
        errors = []
        dns_cache_hash = hash(f"{query}:PTR")
        while tries_left > 0:
            try:
                if dns_cache_hash in self._dns_cache:
                    result = self._dns_cache[dns_cache_hash]
                else:
                    result = list(self._catch(self.resolver.resolve_address, query, **kwargs))
                    if cache_result:
                        self._dns_cache[dns_cache_hash] = result
                return result, errors
            except (dns.exception.Timeout, dns.resolver.LifetimeTimeout, dns.resolver.NoNameservers) as e:
                errors.append(e)
                # don't retry if we get a SERVFAIL
                if isinstance(e, dns.resolver.NoNameservers):
                    self.debug(f"{e} (query={query}, kwargs={kwargs})")
                    break
                else:
                    tries_left -= 1
                    if tries_left > 0:
                        retry_num = (retries + 2) - tries_left
                        self.debug(f"Retrying (#{retry_num}) {query} with kwargs={kwargs}")
        self.debug(f"Results for {query} with kwargs={kwargs}: {results}")
        return results, errors

    def resolve_event(self, event, minimal=False):
        result = self._resolve_event(event, minimal=minimal)
        # if it's a wildcard, go again with _wildcard.{domain}
        if len(result) == 2:
            event, wildcard_rdtypes = result
            return self._resolve_event(event, minimal=minimal, _wildcard_rdtypes=wildcard_rdtypes)
        # else we're good
        else:
            return result

    def _resolve_event(self, event, minimal=False, _wildcard_rdtypes=None):
        """
        Tag event with appropriate dns record types
        Optionally create child events from dns resolutions
        """
        event_tags = set()
        if not event.host or event.type in ("IP_RANGE",):
            return [], set(), False, False, set()
        children = []
        event_host = str(event.host)

        event_whitelisted = False
        event_blacklisted = False

        resolved_hosts = set()

        # wildcard checks
        if not is_ip(event.host):
            # check if this domain is using wildcard dns
            event_in_scope = event.scope_distance == 0
            for hostname, wildcard_domain_rdtypes in self.is_wildcard_domain(
                event_host, log_info=event_in_scope
            ).items():
                if wildcard_domain_rdtypes:
                    event_tags.add("wildcard-domain")
                    for rdtype, ips in wildcard_domain_rdtypes.items():
                        event_tags.add(f"{rdtype.lower()}-wildcard-domain")
            # check if the dns name itself is a wildcard entry
            if _wildcard_rdtypes is None:
                wildcard_rdtypes = self.is_wildcard(event_host)
            else:
                wildcard_rdtypes = _wildcard_rdtypes
            for rdtype, (is_wildcard, wildcard_host) in wildcard_rdtypes.items():
                wildcard_tag = "error"
                if is_wildcard == True:
                    event_tags.add("wildcard")
                    wildcard_tag = "wildcard"
                event_tags.add(f"{rdtype.lower()}-{wildcard_tag}")

        # lock to ensure resolution of the same host doesn't start while we're working here
        with self._event_cache_locks.get_lock(event_host):
            # try to get data from cache
            _event_tags, _event_whitelisted, _event_blacklisted, _resolved_hosts = self.event_cache_get(event_host)
            event_tags.update(_event_tags)
            # if we found it, return it
            if _event_whitelisted is not None:
                return children, event_tags, _event_whitelisted, _event_blacklisted, _resolved_hosts

            # then resolve
            if event.type == "DNS_NAME" and not minimal:
                types = self.all_rdtypes
            else:
                types = ("A", "AAAA")

            futures = {}
            for t in types:
                future = self.submit_task(
                    self._catch_keyboardinterrupt, self.resolve_raw, event_host, type=t, cache_result=True
                )
                if future is None:
                    break
                futures[future] = t

            for future in self.parent_helper.as_completed(futures):
                resolved_raw, errors = future.result()
                for rdtype, e in errors:
                    event_tags.add(f"{rdtype.lower()}-error")
                for rdtype, records in resolved_raw:
                    event_tags.add("resolved")
                    rdtype = str(rdtype).upper()
                    event_tags.add(f"{rdtype.lower()}-record")

                    # whitelisting and blacklisting of IPs
                    for r in records:
                        for _, t in self.extract_targets(r):
                            if t:
                                if rdtype in ("A", "AAAA", "CNAME"):
                                    ip = self.parent_helper.make_ip_type(t)

                                    with suppress(ValidationError):
                                        if self.parent_helper.is_ip(ip):
                                            if self.parent_helper.scan.whitelisted(ip):
                                                event_whitelisted = True
                                    with suppress(ValidationError):
                                        if self.parent_helper.scan.blacklisted(ip):
                                            event_blacklisted = True
                                    resolved_hosts.add(ip)

                                if self.filter_bad_ptrs and rdtype in ("PTR") and self.bad_ptr_regex.search(t):
                                    self.debug(f"Filtering out bad PTR: {t}")
                                    continue
                                children.append((rdtype, t))

            # wildcard event modification (www.evilcorp.com --> _wildcard.evilcorp.com)
            if not is_ip(event.host) and children and _wildcard_rdtypes is None:
                # these are the rdtypes that successfully resolve
                resolved_rdtypes = set([c[0].upper() for c in children])
                # these are the rdtypes that have wildcards
                wildcard_rdtypes_set = set(wildcard_rdtypes)
                # consider the event a full wildcard if all its records are wildcards
                event_is_wildcard = all(r in wildcard_rdtypes_set for r in resolved_rdtypes)
                if event_is_wildcard and event.type in ("DNS_NAME",) and not "_wildcard" in event.data.split("."):
                    wildcard_parent = self.parent_helper.parent_domain(event_host)
                    for rdtype, (_is_wildcard, _parent_domain) in wildcard_rdtypes.items():
                        if _is_wildcard:
                            wildcard_parent = _parent_domain
                            break
                    wildcard_data = f"_wildcard.{wildcard_parent}"
                    if wildcard_data != event.data:
                        log.debug(f'Wildcard detected, changing event.data "{event.data}" --> "{wildcard_data}"')
                        event.data = wildcard_data
                    return (event, wildcard_rdtypes)

            if not self.parent_helper.in_tests:
                ips = set()
                if event.type == "IP_ADDRESS":
                    ips.add(event.data)
                for rdtype, target in children:
                    if rdtype in ("A", "AAAA"):
                        ips.add(target)
                for ip in ips:
                    provider, provider_type, subnet = cloudcheck.check(ip)
                    if provider:
                        event_tags.add(f"{provider_type}-{provider}")

            if "resolved" not in event_tags:
                event_tags.add("unresolved")
            for ip in resolved_hosts:
                try:
                    ip = ipaddress.ip_address(ip)
                    if ip.is_private:
                        event_tags.add("private-ip")
                except ValueError:
                    continue

            self._event_cache[event_host] = (event_tags, event_whitelisted, event_blacklisted, resolved_hosts)

        return children, event_tags, event_whitelisted, event_blacklisted, resolved_hosts

    def event_cache_get(self, host):
        try:
            return self._event_cache[host]
        except KeyError:
            return set(), None, None, set()

    def resolve_batch(self, queries, **kwargs):
        """
        resolve_batch("www.evilcorp.com", "evilcorp.com") --> [
            ("www.evilcorp.com", {"1.1.1.1"}),
            ("evilcorp.com", {"2.2.2.2"})
        ]
        """
        futures = dict()
        for query in queries:
            future = self.submit_task(self._catch_keyboardinterrupt, self.resolve, query, **kwargs)
            if future is None:
                break
            futures[future] = query
        for future in self.parent_helper.as_completed(futures):
            query = futures[future]
            yield (query, future.result())

    def extract_targets(self, record):
        """
        Extract whatever hostnames/IPs a DNS records points to
        """
        results = set()
        rdtype = str(record.rdtype.name).upper()
        if rdtype in ("A", "AAAA", "NS", "CNAME", "PTR"):
            results.add((rdtype, self._clean_dns_record(record)))
        elif rdtype == "SOA":
            results.add((rdtype, self._clean_dns_record(record.mname)))
        elif rdtype == "MX":
            results.add((rdtype, self._clean_dns_record(record.exchange)))
        elif rdtype == "SRV":
            results.add((rdtype, self._clean_dns_record(record.target)))
        elif rdtype == "TXT":
            for s in record.strings:
                s = self.parent_helper.smart_decode(s)
                for match in dns_name_regex.finditer(s):
                    start, end = match.span()
                    host = s[start:end]
                    results.add((rdtype, host))
        else:
            log.warning(f'Unknown DNS record type "{rdtype}"')
        return results

    @staticmethod
    def _clean_dns_record(record):
        with suppress(Exception):
            record = str(record.to_text())
        return str(record).rstrip(".").lower()

    def get_valid_resolvers(self, min_reliability=0.99):
        nameservers = set()
        nameservers_file = self.parent_helper.download(self.nameservers_url, cache_hrs=72)
        if nameservers_file is None:
            log.warning(f"Failed to download nameservers from {self.nameservers_url}")
        else:
            nameservers_json = []
            try:
                nameservers_json = json.loads(open(nameservers_file).read())
            except Exception as e:
                log.warning(f"Failed to load nameserver list from {nameservers_file}: {e}")
                nameservers_file.unlink()
            for entry in nameservers_json:
                try:
                    ip = str(entry.get("ip", "")).strip()
                except Exception:
                    continue
                try:
                    reliability = float(entry.get("reliability", 0))
                except ValueError:
                    continue
                if reliability >= min_reliability and is_ip(ip, version=4):
                    nameservers.add(ip)
            log.verbose(f"Loaded {len(nameservers):,} nameservers from {self.nameservers_url}")
        if not nameservers:
            log.info(f"Loading fallback nameservers from {self.fallback_nameservers_file}")
            lines = self.parent_helper.read_file(self.fallback_nameservers_file)
            nameservers = set([l for l in lines if not l.startswith("#")])
        resolver_list = self.verify_nameservers(nameservers)
        return resolver_list

    def _catch(self, callback, *args, **kwargs):
        try:
            return callback(*args, **kwargs)
        except dns.resolver.NoNameservers:
            raise
        except (dns.exception.Timeout, dns.resolver.LifetimeTimeout):
            log.debug(f"DNS query with args={args}, kwargs={kwargs} timed out after {self.timeout} seconds")
            raise
        except dns.exception.DNSException as e:
            self.debug(f"{e} (args={args}, kwargs={kwargs})")
        except Exception:
            log.warning(f"Error in {callback.__qualname__}() with args={args}, kwargs={kwargs}")
        return list()

    def is_wildcard(self, query, ips=None, rdtype=None):
        """
        Use this method to check whether a *host* is a wildcard entry

        This can reliably tell the difference between a valid DNS record and a wildcard inside a wildcard domain.

        If you want to know whether a domain is using wildcard DNS, use is_wildcard_domain() instead.

        Returns a dictionary in the following format:
            {rdtype: (is_wildcard, wildcard_parent)}

            is_wildcard("www.github.io") --> {"A": (True, "github.io"), "AAAA": (True, "github.io")}

        Note that is_wildcard can be True, False, or None (indicating that wildcard detection was inconclusive)
        """
        result = {}
        if rdtype is None:
            rdtype = "ANY"

        query = self._clean_dns_record(query)
        # skip check if it's an IP
        if is_ip(query) or not "." in query:
            return {}
        # skip check if the query is a domain
        if is_domain(query):
            return {}
        # skip check if the query's parent domain is excluded in the config
        for d in self.wildcard_ignore:
            if self.parent_helper.host_in_host(query, d):
                return {}

        parent = parent_domain(query)
        parents = list(domain_parents(query))

        futures = []
        base_query_ips = dict()
        # if the caller hasn't already done the work of resolving the IPs
        if ips is None:
            # then resolve the query for all rdtypes
            for _rdtype in self.all_rdtypes:
                # resolve the base query
                future = self.submit_task(
                    self._catch_keyboardinterrupt, self.resolve_raw, query, type=_rdtype, cache_result=True
                )
                if future is None:
                    break
                futures.append(future)

            for future in self.parent_helper.as_completed(futures):
                raw_results, errors = future.result()
                if errors and not raw_results:
                    self.debug(f"Failed to resolve {query} ({_rdtype}) during wildcard detection")
                    result[_rdtype] = (None, parent)
                    continue
                for _rdtype, answers in raw_results:
                    base_query_ips[_rdtype] = set()
                    for answer in answers:
                        for _, t in self.extract_targets(answer):
                            base_query_ips[_rdtype].add(t)
        else:
            # otherwise, we can skip all that
            base_query_ips[rdtype] = set([self._clean_dns_record(ip) for ip in ips])
        if not base_query_ips:
            return result

        # once we've resolved the base query and have IP addresses to work with
        # we can compare the IPs to the ones we have on file for wildcards
        # for every rdtype
        for _rdtype in self.all_rdtypes:
            # get the IPs from above
            query_ips = base_query_ips.get("ANY", base_query_ips.get(_rdtype, set()))
            if not query_ips:
                continue
            # for every parent domain, starting with the longest
            for host in parents[::-1]:
                host_hash = hash(host)
                # make sure we've checked that domain for wildcards
                self.is_wildcard_domain(host)
                if host_hash in self._wildcard_cache:
                    # then get its IPs from our wildcard cache
                    wildcard_rdtypes = self._wildcard_cache[host_hash]
                    # then check to see if our IPs match the wildcard ones
                    if _rdtype in wildcard_rdtypes:
                        wildcard_ips = wildcard_rdtypes[_rdtype]
                        # if our IPs match the wildcard ones, then ladies and gentlemen we have a wildcard
                        is_wildcard = any(r in wildcard_ips for r in query_ips)
                        if is_wildcard:
                            result[_rdtype] = (True, host)
                            break

        return result

    def is_wildcard_domain(self, domain, log_info=False):
        """
        Check whether a domain is using wildcard DNS

        Returns a dictionary containing any DNS record types that are wildcards, and their associated IPs
            is_wildcard_domain("github.io") --> {"A": {"1.2.3.4",}, "AAAA": {"dead::beef",}}
        """
        wildcard_domain_results = {}
        domain = self._clean_dns_record(domain)

        # make a list of its parents
        parents = list(domain_parents(domain, include_self=True))
        # and check each of them, beginning with the highest parent (i.e. the root domain)
        for i, host in enumerate(parents[::-1]):
            # have we checked this host before?
            host_hash = hash(host)
            with self._wildcard_lock.get_lock(host_hash):
                # if we've seen this host before
                if host_hash in self._wildcard_cache:
                    wildcard_domain_results[host] = self._wildcard_cache[host_hash]
                    continue

                # determine if this is a wildcard domain
                wildcard_futures = {}
                # resolve a bunch of random subdomains of the same parent
                for rdtype in self.all_rdtypes:
                    # continue if a wildcard was already found for this rdtype
                    # if rdtype in self._wildcard_cache[host_hash]:
                    #     continue
                    for _ in range(self.wildcard_tests):
                        rand_query = f"{rand_string(digits=False, length=10)}.{host}"
                        future = self.submit_task(
                            self._catch_keyboardinterrupt,
                            self.resolve,
                            rand_query,
                            type=rdtype,
                            cache_result=False,
                        )
                        if future is None:
                            break
                        wildcard_futures[future] = rdtype

                # combine the random results
                is_wildcard = False
                wildcard_results = dict()
                for future in self.parent_helper.as_completed(wildcard_futures):
                    results = future.result()
                    rdtype = wildcard_futures[future]
                    if results:
                        is_wildcard = True
                    if results:
                        if not rdtype in wildcard_results:
                            wildcard_results[rdtype] = set()
                        wildcard_results[rdtype].update(results)

                self._wildcard_cache.update({host_hash: wildcard_results})
                wildcard_domain_results.update({host: wildcard_results})
                if is_wildcard:
                    wildcard_rdtypes_str = ",".join(sorted([t.upper() for t, r in wildcard_results.items() if r]))
                    log_fn = log.verbose
                    if log_info:
                        log_fn = log.info
                    log_fn(f"Encountered domain with wildcard DNS ({wildcard_rdtypes_str}): {host}")

        return wildcard_domain_results

    def _catch_keyboardinterrupt(self, callback, *args, **kwargs):
        try:
            return callback(*args, **kwargs)
        except Exception as e:
            log.error(f"Error in {callback.__qualname__}(): {e}")
            log.trace(traceback.format_exc())
        except KeyboardInterrupt:
            if self.parent_helper.scan:
                self.parent_helper.scan.stop()

    def debug(self, *args, **kwargs):
        if self._debug:
            log.debug(*args, **kwargs)

    def _get_dummy_module(self, name):
        with self._dummy_modules_lock:
            try:
                dummy_module = self._dummy_modules[name]
            except KeyError:
                dummy_module = self.parent_helper._make_dummy_module(name=name, _type="DNS")
                self._dummy_modules[name] = dummy_module
        return dummy_module

    def dns_warning(self, msg):
        msg_hash = hash(msg)
        if msg_hash not in self._dns_warnings:
            log.warning(msg)
            self._dns_warnings.add(msg_hash)



import ast
import sys
import importlib
from pathlib import Path
from omegaconf import OmegaConf
from contextlib import suppress

from .misc import list_files, sha1, search_dict_by_key, search_format_dict, make_table, os_platform


class ModuleLoader:
    def __init__(self):
        self._preloaded = {}
        self._preloaded_orig = None
        self._modules = {}
        self._configs = {}

    def file_filter(self, file):
        return file.suffix.lower() == ".py" and file.stem not in ["base", "__init__"]

    def preload(self, module_dir):
        """
        Preload modules from a specified directory
        """
        module_dir = Path(module_dir)
        for module_file in list_files(module_dir, filter=self.file_filter):
            if module_dir.name == "modules":
                namespace = f"bbot.modules"
            else:
                namespace = f"bbot.modules.{module_dir.name}"
            try:
                preloaded = self.preload_module(module_file)
                module_type = "scan"
                if module_dir.name in ("output", "internal"):
                    module_type = str(module_dir.name)
                elif module_dir.name not in ("modules"):
                    preloaded["flags"] = list(set(preloaded["flags"] + [module_dir.name]))
                preloaded["type"] = module_type
                preloaded["namespace"] = namespace
                config = OmegaConf.create(preloaded.get("config", {}))
                self._configs[module_file.stem] = config
                self._preloaded[module_file.stem] = preloaded
            except Exception:
                import traceback

                print(f"[CRIT] Error preloading {module_file}\n\n{traceback.format_exc()}")
                print(f"[CRIT] Error in {module_file.name}")
                sys.exit(1)

        return self.preloaded

    def preloaded(self, type=None):
        preloaded = {}
        if type is not None:
            preloaded = {k: v for k, v in self._preloaded.items() if self.check_type(k, type)}
        else:
            preloaded = dict(self._preloaded)
        return preloaded

    def configs(self, type=None):
        configs = {}
        if type is not None:
            configs = {k: v for k, v in self._configs.items() if self.check_type(k, type)}
        else:
            configs = dict(self._configs)
        return OmegaConf.create(configs)

    def find_and_replace(self, **kwargs):
        if self._preloaded_orig is None:
            self._preloaded_orig = dict(self._preloaded)
        self._preloaded = search_format_dict(self._preloaded_orig, **kwargs)

    def check_type(self, module, type):
        return self._preloaded[module]["type"] == type

    def preload_module(self, module_file):
        watched_events = []
        produced_events = []
        flags = []
        meta = {}
        pip_deps = []
        shell_deps = []
        apt_deps = []
        ansible_tasks = []
        python_code = open(module_file).read()
        # take a hash of the code so we can keep track of when it changes
        module_hash = sha1(python_code).hexdigest()
        parsed_code = ast.parse(python_code)
        config = {}
        options_desc = {}
        for root_element in parsed_code.body:
            # look for classes
            if type(root_element) == ast.ClassDef:
                for class_attr in root_element.body:
                    # class attributes that are dictionaries
                    if type(class_attr) == ast.Assign and type(class_attr.value) == ast.Dict:
                        # module options
                        if any([target.id == "options" for target in class_attr.targets]):
                            config.update(ast.literal_eval(class_attr.value))
                        # module options
                        if any([target.id == "options_desc" for target in class_attr.targets]):
                            options_desc.update(ast.literal_eval(class_attr.value))
                        # module metadata
                        if any([target.id == "meta" for target in class_attr.targets]):
                            meta = ast.literal_eval(class_attr.value)
                    # class attributes that are lists
                    if type(class_attr) == ast.Assign and type(class_attr.value) == ast.List:
                        # flags
                        if any([target.id == "flags" for target in class_attr.targets]):
                            for flag in class_attr.value.elts:
                                if type(flag.value) == str:
                                    flags.append(flag.value)
                        # watched events
                        if any([target.id == "watched_events" for target in class_attr.targets]):
                            for event_type in class_attr.value.elts:
                                if type(event_type.value) == str:
                                    watched_events.append(event_type.value)
                        # produced events
                        if any([target.id == "produced_events" for target in class_attr.targets]):
                            for event_type in class_attr.value.elts:
                                if type(event_type.value) == str:
                                    produced_events.append(event_type.value)
                        # python dependencies
                        if any([target.id == "deps_pip" for target in class_attr.targets]):
                            for python_dep in class_attr.value.elts:
                                if type(python_dep.value) == str:
                                    pip_deps.append(python_dep.value)
                        # apt dependencies
                        elif any([target.id == "deps_apt" for target in class_attr.targets]):
                            for apt_dep in class_attr.value.elts:
                                if type(apt_dep.value) == str:
                                    apt_deps.append(apt_dep.value)
                        # bash dependencies
                        elif any([target.id == "deps_shell" for target in class_attr.targets]):
                            for shell_dep in class_attr.value.elts:
                                shell_deps.append(ast.literal_eval(shell_dep))
                        # ansible playbook
                        elif any([target.id == "deps_ansible" for target in class_attr.targets]):
                            ansible_tasks = ast.literal_eval(class_attr.value)
        for task in ansible_tasks:
            if not "become" in task:
                task["become"] = False
            # don't sudo brew
            elif os_platform() == "darwin" and ("package" in task and task.get("become", False) == True):
                task["become"] = False
        preloaded_data = {
            "watched_events": watched_events,
            "produced_events": produced_events,
            "flags": flags,
            "meta": meta,
            "config": config,
            "options_desc": options_desc,
            "hash": module_hash,
            "deps": {"pip": pip_deps, "shell": shell_deps, "apt": apt_deps, "ansible": ansible_tasks},
            "sudo": len(apt_deps) > 0,
        }
        if any(x == True for x in search_dict_by_key("become", ansible_tasks)) or any(
            x == True for x in search_dict_by_key("ansible_become", ansible_tasks)
        ):
            preloaded_data["sudo"] = True
        return preloaded_data

    def load_modules(self, module_names):
        modules = {}
        for module_name in module_names:
            module = self.load_module(module_name)
            modules[module_name] = module
            self._modules[module_name] = module
        return modules

    def load_module(self, module_name):
        namespace = self._preloaded[module_name]["namespace"]
        import_path = f"{namespace}.{module_name}"
        module_variables = importlib.import_module(import_path, "bbot")
        # for every top-level variable in the .py file
        for variable in module_variables.__dict__.keys():
            # get its value
            value = getattr(module_variables, variable)
            with suppress(AttributeError):
                # if it has watched_events and produced_events
                if all(
                    type(a) == list
                    for a in (getattr(value, "watched_events", None), getattr(value, "produced_events", None))
                ):
                    # and if its variable name matches its filename
                    if value.__name__.lower() == module_name.lower():
                        value._name = module_name
                        # then we have a module
                        return value

    def recommend_dependencies(self, modules):
        """
        Returns a dictionary containing missing dependencies and their suggested resolutions
        """
        resolve_choices = {}
        # step 1: build a dictionary containing event types and their associated modules
        # {"IP_ADDRESS": set("naabu", "ipneighbor", ...)}
        watched = {}
        produced = {}
        for modname in modules:
            preloaded = self._preloaded.get(modname)
            if preloaded:
                for event_type in preloaded.get("watched_events", []):
                    self.add_or_create(watched, event_type, modname)
                for event_type in preloaded.get("produced_events", []):
                    self.add_or_create(produced, event_type, modname)
        watched_all = {}
        produced_all = {}
        for modname, preloaded in self.preloaded().items():
            if preloaded:
                for event_type in preloaded.get("watched_events", []):
                    self.add_or_create(watched_all, event_type, modname)
                for event_type in preloaded.get("produced_events", []):
                    self.add_or_create(produced_all, event_type, modname)

        # step 2: check to see if there are missing dependencies
        for modname in modules:
            preloaded = self._preloaded.get(modname)
            module_type = preloaded.get("type", "unknown")
            if module_type != "scan":
                continue
            watched_events = preloaded.get("watched_events", [])
            missing_deps = {e: not self.check_dependency(e, modname, produced) for e in watched_events}
            if all(missing_deps.values()):
                for event_type in watched_events:
                    if event_type == "SCAN":
                        continue
                    choices = produced_all.get(event_type, [])
                    choices = set(choices)
                    with suppress(KeyError):
                        choices.remove(modname)
                    if event_type not in resolve_choices:
                        resolve_choices[event_type] = dict()
                    deps = resolve_choices[event_type]
                    self.add_or_create(deps, "required_by", modname)
                    for c in choices:
                        choice_type = self._preloaded.get(c, {}).get("type", "unknown")
                        if choice_type == "scan":
                            self.add_or_create(deps, "recommended", c)

        return resolve_choices

    def check_dependency(self, event_type, modname, produced):
        if event_type not in produced:
            return False
        if produced[event_type] == {modname}:
            return False
        return True

    @staticmethod
    def add_or_create(d, k, *items):
        try:
            d[k].update(set(items))
        except KeyError:
            d[k] = set(items)

    def modules_table(self, modules=None, mod_type=None):
        table = []
        header = ["Module", "Type", "Needs\nAPI\nKey", "Description", "Flags", "Produced Events"]
        maxcolwidths = [20, 20, 5, 40, 40, 40]
        for module_name, preloaded in self.filter_modules(modules, mod_type):
            module_type = preloaded["type"]
            produced_events = sorted(preloaded.get("produced_events", []))
            flags = sorted(preloaded.get("flags", []))
            api_key_required = ""
            meta = preloaded.get("meta", {})
            if meta.get("auth_required", False):
                api_key_required = "X"
            description = meta.get("description", "")
            table.append(
                [module_name, module_type, api_key_required, description, ",".join(flags), ",".join(produced_events)]
            )
        return make_table(table, header, maxcolwidths=maxcolwidths)

    def modules_options_table(self, modules=None, mod_type=None):
        table = []
        header = ["Option", "Type", "Default", "Description"]
        for module_name, preloaded in self.filter_modules(modules, mod_type):
            module_type = preloaded["type"]
            module_options = preloaded["config"]
            module_options_desc = preloaded["options_desc"]
            for k, v in module_options.items():
                if module_type not in ("internal", "output"):
                    module_key = "modules"
                else:
                    module_key = f"{module_type}_modules"
                option_name = f"{module_key}.{module_name}.{k}"
                option_type = type(v).__name__
                option_description = module_options_desc[k]
                table.append([option_name, option_type, str(v), option_description])
        return make_table(table, header)

    def filter_modules(self, modules=None, mod_type=None):
        if modules is None:
            module_list = list(self.preloaded(type=mod_type).items())
        else:
            module_list = [(m, self._preloaded[m]) for m in modules]
        module_list.sort(key=lambda x: x[0])
        module_list.sort(key=lambda x: "passive" in x[-1]["flags"])
        module_list.sort(key=lambda x: x[-1]["type"], reverse=True)
        return module_list


module_loader = ModuleLoader()



from .crobat import crobat


class anubisdb(crobat):
    flags = ["subdomain-enum", "passive", "safe"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {"description": "Query jldc.me's database for subdomains"}

    base_url = "https://jldc.me/anubis/subdomains"
    dns_abort_depth = 5

    def request_url(self, query):
        url = f"{self.base_url}/{self.helpers.quote(query)}"
        return self.request_with_fail_count(url)

    def abort_if_pre(self, hostname):
        """
        Discards results that are longer than 5 segments, e.g. a.b.c.d.evilcorp.com
        This exists because of the _disgusting_ amount of garbage data in this API
        """
        dns_depth = hostname.count(".") + 1
        if dns_depth > self.dns_abort_depth:
            return True
        return False

    def abort_if(self, event):
        # abort if dns name is unresolved
        return (not "resolved" in event.tags) or super().abort_if(event)

    def parse_results(self, r, query):
        results = set()
        json = r.json()
        if json:
            for hostname in json:
                hostname = str(hostname).lower()
                if hostname.endswith(f".{query}") and not self.abort_if_pre(hostname):
                    results.add(hostname)
        return results
